#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <recarray.h>
#include <utility.h>

#include "soggetti.h"
#include "benem.h"
#include "sezioni.h"

#include "at1.h"
#include "at1100a.h"

#define ALIAS_LCP	100
#define ALIAS_TCS	200

enum ts { undefined = 0, elenco = 1, etichette = 2 };

// definizione form per etichette
class TEti_donben_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TEti_donben_form(): TForm() {};
  TEti_donben_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TEti_donben_form() {};
};

class TStampaDonBen : public TPrintapp
{
  TRelation*					_rel;
  TMask*      				_msk;
  TLocalisamfile*			_benem;
	TRecord_array*  		_sbenemerenze;
  TEti_donben_form*		_form_eti;
  TAssoc_array		    _categorie;
  int 								_cur;
  TParagraph_string 	_cognome_nome;
  TDate 							_data_stampa;
	ts 	      					_tipostampa;
	TString16						_codsez, _codsot;
	int 								_numini, _numfin;
	TString16 					_tipoben;
	bool								_definitiva;
	int 								_contatore;
  int									_etlarghezza, _etcolonne;

  static bool filter_func_donben(const TRelation* rel);
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page(int file, int counter);
  virtual print_action postprocess_print(int file, int counter);

public:
  void crea_intestazione();
  void filtra_sezioni();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void footer_sezione();
  TMask& app_mask() { return *_msk; }
  TStampaDonBen() : _data_stampa(TODAY), _cognome_nome("",25) {}
};

HIDDEN inline TStampaDonBen& app() { return (TStampaDonBen&) main_app(); }

TCursor* TEti_donben_form::cursor() const { return app().current_cursor(); }

TRelation* TEti_donben_form::relation() const { return cursor()->relation(); }

print_action TStampaDonBen::postprocess_print(int file, int counter)
{ 
	if (_tipostampa == elenco && _contatore > 0)
		footer_sezione();
	return NEXT_PAGE;
}

void TStampaDonBen::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  TRectype da(LF_SOGGETTI);
  TRectype a (LF_SOGGETTI);
  if (sezini.not_empty())
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

void TStampaDonBen::set_page(int file, int cnt)
{
	// costruzione etichette
	switch (_tipostampa)
	{   
		case etichette:
		{ 
			TPrint_section& corpo = _form_eti->get_body();
			for (int r=1;r<=_etcolonne;r++)
			{                     
				if (current_cursor()->pos()<current_cursor()->items())
				{ 
					force_setpage(TRUE);
					corpo.update();
  				for (word i = 0; i < corpo.height(); i++)
  				{
  					TPrintrow& riga = corpo.row(i);
  					TString256 riga1 = riga.row();
						riga1.cut(_etlarghezza);
  					int colonna = ((r-1)*_etlarghezza);
  					TString16 formato;
  					formato << '@' << colonna << "g";
  					riga1.insert(formato,0);
  					set_row(i+1,riga1);
					}
					if (r < _etcolonne)  			
						++(*current_cursor());
				}					
			}				
		}
		break;
		case elenco:
		{
			if (_definitiva)
				set_row(1,"@5g#a", &_cognome_nome);
			else
			{
				set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
				set_row(2,"@0g@8,rs", FLD(LF_SOGGETTI,SOG_TESSAVIS));
				set_row(1,"@9g@S", FLD(LF_SOGGETTI,SOG_CATDON));
				set_row(2,"@9g@S", FLD(LF_SOGGETTI,SOG_CODSEZ));
				set_row(2,"@12g@S", FLD(LF_SOGGETTI,SOG_CODSOT));
				set_row(1,"@15g#a", &_cognome_nome);
				set_row(2,"@15g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
				set_row(1,"@41g@pn", FLD(LF_SOGGETTI,SOG_TOTDON,"###"));
				set_row(1,"@45g@ld", FLD(LF_SOGGETTI,SOG_DATAULTDON));
				set_row(2,"@48g@S", FLD(LF_SOGGETTI,SOG_TIPOULTDON));
				set_row(1,"@56g@S", FLD(LF_SOGGETTI,SOG_STATO));
				set_row(1,"@59g@ld", FLD(LF_SOGGETTI,SOG_DATAPROSSI));
				set_row(2,"@59g@ld", FLD(LF_SOGGETTI,SOG_DATAPROSAF));
				set_row(3,"");
			}				
		}
		break;
	}					
}

bool TStampaDonBen::filter_func_donben(const TRelation* rel)
{   
	bool filtrato = TRUE;
	TLocalisamfile& sog = rel->lfile(); 
	// filtro per categorie
	TAssoc_array& categorie = app()._categorie;
	long codice = sog.get_long(SOG_CODICE);
	if (categorie.items() != 0)
	{
		const TString16 cat = sog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) cat);
	}				
	// filtro per numero donazioni
	if (filtrato)
	{
		const int totdon = sog.get_int(SOG_TOTDON);
		filtrato = ((totdon >= app()._numini) && (totdon <= app()._numfin));
	}
	// filtro per tipo benemerenza
	if ((filtrato) && (app()._tipoben.not_empty()))
	{
		TLocalisamfile benem(LF_BENEM);
		benem.setkey(3);
		benem.zero();
		benem.put(BEN_CODICE, codice);
		benem.put(BEN_TIPOBEN, app()._tipoben);
		int err = benem.read();
		filtrato = (err != NOERR);
	}	
	if (filtrato && app()._definitiva)
	{
		TRectype* key = new TRectype(LF_BENEM);
		key->put(BEN_CODICE, codice);
		int err = app()._sbenemerenze->read(key);
		if (err == NOERR)
		{                                    
			int r = app()._sbenemerenze->rows();
			TRectype& rec = app()._sbenemerenze->row(r+1,TRUE);
			
			TString16 codsez = sog.get(SOG_CODSEZ);
			TString16 codsot = sog.get(SOG_CODSOT);
			TString16 gruppoazie = sog.get(SOG_GRUPPOAZIE);
						
			rec.put(BEN_TIPOBEN, 		app()._tipoben);
			//rec.put(BEN_DATAMAT, 		row.get(2));
			//rec.put(BEN_BONUS, 			row.get()); 
			TDate oggi(TODAY);
			rec.put(BEN_DATABEN, 		oggi);
			//rec.put(BEN_DATACON, 		row.get());
			//rec.put(BEN_RITIRATA, 	row.get());
			rec.put(BEN_CODSEZ, 		codsez);
			rec.put(BEN_CODSOT, 		codsot);
			rec.put(BEN_GRUPPOAZIE,	gruppoazie);
			int err = app()._sbenemerenze->rewrite();
		}
	}		
  return filtrato;
}

bool TStampaDonBen::preprocess_page(int file, int counter)
{
	if (_tipostampa == elenco)
	{
		TRectype& recsog = current_cursor()->curr();
		TString80 nome = recsog.get(SOG_COGNOME);
		nome << " ";
		nome << recsog.get(SOG_NOME);
		_cognome_nome = nome;
		// salto pagina se cambio sezione
		const TString16 codsez = recsog.get(SOG_CODSEZ);		
		const TString16 codsot = recsog.get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if ((_codsez != "**") && (_contatore > 0))
				footer_sezione();
			_contatore = 0;				
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
		else
		if (printer().rows_left() < 4)
			printer().formfeed();
		_contatore++;			
	}		
	if (_tipostampa==etichette)		
		if (printer().rows_left() < _form_eti->get_body().height())
			printer().formfeed();
  return TRUE;
}

void TStampaDonBen::footer_sezione()
{
	reset_footer();
	TString sep(80);
	sep.fill('-');
	set_footer(2, (const char *) sep);
	set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _contatore);
	printer().formfeed(); 
	reset_footer();
}

void TStampaDonBen::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(80);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if ((densot.ok())&& (densot.not_empty()))
	{
		intestazione << "/";
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

bool TStampaDonBen::set_print(int m)
{ 
	_tipostampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  	_tipostampa = elenco;
  	_codsez = "**";
  	_codsot = "**";
  	break;
  case F_ETICHETTE:
  	_tipostampa = etichette;	
  	break;
  }	
	if (_tipostampa != undefined)
  {
    reset_files(); 
    add_file(LF_SOGGETTI);
    // filtro per sezioni selezionati
	  filtra_sezioni();
	  // filtro per categorie
  	const TString16 catpri = _msk->get(F_CAT1);
  	const TString16 catsec = _msk->get(F_CAT2);
  	const TString16 catter = _msk->get(F_CAT3);
  	const TString16 catqua = _msk->get(F_CAT4);
  	const TString16 catqui = _msk->get(F_CAT5);
  	const TString16 catses = _msk->get(F_CAT6);
	  if (catpri.not_empty() && catpri.ok())
  		_categorie.add((const char*) catpri);
  	if (catsec.not_empty() && catsec.ok())
  		_categorie.add((const char*) catsec);
  	if (catter.not_empty() && catter.ok())
  		_categorie.add((const char*) catter);
  	if (catqua.not_empty() && catqua.ok())
  		_categorie.add((const char*) catqua);
  	if (catqui.not_empty() && catqui.ok())
  		_categorie.add((const char*) catqui);
  	if (catses.not_empty() && catses.ok())
  		_categorie.add((const char*) catses);
		_numini = _msk->get_int(F_NUMINI);  		
		_numfin = _msk->get_int(F_NUMFIN);		
		_tipoben = _msk->get(F_TIPOBEN);
		_definitiva = _msk->get_bool(F_DEFINITIVA);
		if (_definitiva)
			_cognome_nome.set_width(60);
		current_cursor()->set_filterfunction(filter_func_donben);
		reset_print();
		crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TStampaDonBen::crea_intestazione()
{
  reset_header();
  
  if (_tipostampa == elenco)
  {
  	if (_definitiva)
		{
	  	TString sep(80);
	  	sep = "BENEMERENZA ";
	  	TString80 benem = _msk->get(F_D_TIPOBEN);
  		sep << benem;
			sep.center_just();
	  	set_header(2, "@0g%s", (const char*) sep);
	  	const TString16 data_stampa = _data_stampa.string();
	  	set_header(2,"@0g%10s", (const char*) data_stampa);
	  	sep = "";
	  	sep << "Pag. @#";
	  	set_header(2, "@72g%s", (const char*) sep);  
	  	sep = "";
	  	set_header(3, (const char *) sep);
		}  	
  	else
  	{
	  	TString sep(80);
	  	sep = "ELENCO SOGGETTI CHE HANNO DA ";
	  	sep << format("%i A %i DONAZIONI",_numini,_numfin);
			sep.center_just();
	  	set_header(2, "@0g%s", (const char*) sep);
	  	const TString16 data_stampa = _data_stampa.string();
	  	set_header(2,"@0g%10s", (const char*) data_stampa);
	  	sep = "";
	  	sep << "Pag. @#";
	  	set_header(2, "@72g%s", (const char*) sep);  
	  	if (_tipoben.not_empty())
	  	{
	  		sep = "CHE NON HANNO LA BENEMERENZA ";
	  		sep << _tipoben;
	  		sep << " ";
		  	TString80 benem = _msk->get(F_D_TIPOBEN);
  			sep << benem;
				sep.center_just();
	  		set_header(3, "@0g%s", (const char*) sep);
			}  		
	  	sep = "";
	  	sep.fill('-');
	  	set_header(4, (const char *) sep);
	  	set_header(5,"@0gCodice@9gCat.@15gCognome e nome@41gDon@45gData ult.@56gSt@59gData pr.SI");
	  	set_header(6,"@0gTessera@9gSe/So@15gData di nascita@45gTipo ult.@59gData pr.AF");
	  	set_header(7,"@0g--------@9g-----@15g-------------------------@41g---@45g----------@56g--@59g----------");
		}	  	
  	printer().footerlen(3);
	}  	
}

bool TStampaDonBen::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCP);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM");
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");
  _cur = add_cursor(new TCursor(_rel, "", 3));
  _msk = new TMask("at1100a");
	_benem 		=	new TLocalisamfile(LF_BENEM);
	_sbenemerenze = new TRecord_array(LF_BENEM,BEN_PROGBEN);	
  TConfig config(CONFIG_STUDIO);
  TString16 etformato = config.get("EtFormato");
  _etlarghezza = config.get_int("EtLarghezza");
  _etcolonne = config.get_int("EtColonne");
	_form_eti = new TEti_donben_form(etformato);
  return TRUE;
}

bool TStampaDonBen::user_destroy()
{
  delete _msk;
  delete _rel;
  delete _form_eti;
  delete _benem;
  delete _sbenemerenze;
  return TRUE;
}

int at1100(int argc, char* argv[])
{
  TStampaDonBen a;
  a.run(argc, argv, "Proposta benemerenza");
  return 0;
}
