#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <utility.h>

#include "benem.h"
#include "soggetti.h"
#include "sezioni.h"
#include <comuni.h>

#include "at1.h"
#include "at1300a.h"

#define ALIAS_BNZ	600

#define ALIAS_LCPRES	101		// localita' postale di residenza
#define ALIAS_LCPDOM	100		// localita' postale di domicilio
#define ALIAS_COMRES 	502		// comune di residenza
#define ALIAS_COMDOM 	501		// comune di domicilio
#define ALIAS_COMNAS 	500		// comune di nascita

class TStampaBenemerenze : public TPrintapp
{
  TRelation*					_rel;
  TMask*      				_msk;
  int									_counter;
  int 								_cur;
  TDate								_dataini, _datafin;
	TString16						_codsez, _codsot;
	bool								_stampa80;  
  TParagraph_string 	_cognome_nome, _dencom;
  TString16						_tipoold;
  int 								_contatore;
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page(int file, int counter);
  virtual print_action postprocess_print(int file, int counter);

public:
  void crea_intestazione();
  void header_ben(const TString16 tipoben);
  void footer_ben();
  TMask& app_mask() { return *_msk; }
  TStampaBenemerenze() : _cognome_nome("",35), _dencom("",50) {}
};

HIDDEN inline TStampaBenemerenze& app() { return (TStampaBenemerenze&) main_app(); }

print_action TStampaBenemerenze::postprocess_print(int file, int counter)
{ 
	if (_contatore > 0)
		footer_ben();
	return NEXT_PAGE;
}

void TStampaBenemerenze::footer_ben()
{
	// stampa totale benemerenze
	reset_footer();
	TString sep(80);
	sep.fill('-');
	set_footer(2, (const char *) sep);
	set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _contatore);
	printer().formfeed(); 
	reset_footer();
}

void TStampaBenemerenze::set_page(int file, int cnt)
{
	set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
	set_row(1,"@9g@S", FLD(LF_SOGGETTI,SOG_CATDON));
	set_row(1,"@12g#a", &_cognome_nome);
	set_row(1,"@38g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
	set_row(1,"@49g@S", FLD(LF_SOGGETTI,SOG_DOM_INDIR));
	set_row(1,"@100g@S", FLD(LF_SOGGETTI,SOG_TELABI));
	set_row(1,"@116g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
	set_row(2,"@49g#a", &_dencom);
	set_row(2,"@116g@S", FLD(LF_SOGGETTI,SOG_TELALT));
}

bool TStampaBenemerenze::preprocess_page(int file, int counter)
{
	TString80 nome = current_cursor()->curr(LF_SOGGETTI).get(SOG_COGNOME);
	nome << " ";
	nome << current_cursor()->curr(LF_SOGGETTI).get(SOG_NOME);
	_cognome_nome = nome;
	TString256 localita = "";
	localita << current_cursor()->curr(-ALIAS_LCPDOM).get("S6");
	if (localita.not_empty() && localita.ok())
	{
		localita << " ";
		localita << current_cursor()->curr(-ALIAS_LCPDOM).get("S0");
		localita << " - ";		
	}
	else
	{
		localita = current_cursor()->curr(-ALIAS_COMDOM).get(COM_CAPCOM);
		localita << " ";
	}			
	localita << current_cursor()->curr(-ALIAS_COMDOM).get(COM_DENCOM);
	localita.trim();
	_dencom = localita;
	
	// salto pagina se cambio punto di rottura
	TString16 tiponew = current_cursor()->curr().get(BEN_TIPOBEN);		
	if (tiponew != _tipoold )
	{
		if (_tipoold != "**")
			footer_ben();
		_contatore = 0;				
		_tipoold = tiponew;
		header_ben(tiponew);
	}
	else
	if (printer().rows_left() < 4)
		printer().formfeed();
	_contatore++;			
  return TRUE;
}

void TStampaBenemerenze::header_ben(const TString16 tipoben)
{
	TString intestazione(132);
	intestazione = "BENEMERENZA ";
	intestazione << tipoben;
	intestazione << " ";
	intestazione << current_cursor()->curr(-ALIAS_BNZ).get("S0");
	intestazione.center_just(132);	
	set_header(2,"@0g%s", (const char*) intestazione);
	return;
}

bool TStampaBenemerenze::set_print(int m)
{ 
  KEY tasto;
 	tasto = _msk->run();
 	if (tasto == K_ENTER)
 	{ 
		_dataini = 	_msk->get_date(F_DATAINI);	
		_datafin = 	_msk->get_date(F_DATAFIN);	
	  TRectype da(LF_BENEM);
  	TRectype a (LF_BENEM);
  	if (_dataini.ok())
			da.put(BEN_DATABEN, _dataini);
  	if (_datafin.ok())
			a.put(BEN_DATABEN, _datafin);
		TString80 chiave = "94->TIPOBEN|UPPER(90->COGNOME)|UPPER(90->NOME)";
	  _cur = add_cursor(new TSorted_cursor(_rel, (const char*) chiave, "", 2, &da, &a));
	  TString80 filtro = "";
		current_cursor()->setfilter((const char*) filtro, TRUE);
 		_counter = 0;
 		_tipoold = "**";
	  reset_files(); 
  	add_file(LF_BENEM);
		reset_print();
		crea_intestazione();
		return TRUE;
  }	            
 	else
  	return FALSE;
}

void TStampaBenemerenze::crea_intestazione()
{
  reset_header();
  TString sep(132);
	sep = "STAMPA BENEMERENZE ATTRIBUITE dal ";
	sep << _dataini;
	sep << " al ";
	sep << _datafin;
	sep.center_just(132);			
	set_header(1,"@0g%s", (const char*) sep);
 	sep = "Pag. @#";
	set_header(1, "@110g%s", (const char*) sep);  
	sep = "";
	sep.fill('-');
	set_header(3, (const char *) sep);
	set_header(4,"@0gCodice@9gC.@12gCognome e nome@38gNato il@49gIndirizzo@100gTelefono abit.@116gTelefono lavoro");
	set_header(5,"@49gCAP/Localit/Comune/Prov.@116gTelefono altro");
	set_header(6,"@0g--------@9g--@12g-------------------------@38g----------@49g--------------------------------------------------@100g---------------@116g---------------");
 	printer().footerlen(3);
}

bool TStampaBenemerenze::user_create()
{
  _rel = new TRelation(LF_BENEM);
  _rel->add(LF_SOGGETTI, "CODICE==CODICE");
  _rel->add("BNZ", "CODTAB==TIPOBEN",1,0,ALIAS_BNZ);
  _rel->add("LCP", "CODTAB==RES_CODLOC",1,LF_SOGGETTI,ALIAS_LCPRES);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,LF_SOGGETTI,ALIAS_LCPDOM);
  _rel->add(LF_COMUNI, "COM==RES_CODCOM",1,LF_SOGGETTI,ALIAS_COMRES);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM",1,LF_SOGGETTI,ALIAS_COMDOM);
  _rel->add(LF_COMUNI, "COM==COMNASC",1,LF_SOGGETTI,ALIAS_COMNAS);
  
  _msk = new TMask("at1300a");
  TConfig config(CONFIG_STUDIO);
  _stampa80 = config.get_bool("Stampa80");
  return TRUE;
}

bool TStampaBenemerenze::user_destroy()
{
  delete _msk;
  delete _rel;
  return TRUE;
}

int at1300(int argc, char* argv[])
{
  TStampaBenemerenze a;
  a.run(argc, argv, "Stampa benemerenze attribuite");
  return 0;
}
