#include <form.h>
#include <mask.h>
#include <printapp.h>

#include "soggetti.h"
#include "sezioni.h"
#include <comuni.h>

#include "at2.h"
#include "at2800a.h"

#define ALIAS_LCP	100
#define ALIAS_TCS	200

enum ts { undefined = 0, elenco = 1, etichette = 2 };

// definizione form per etichette
class TEti_nondon_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TEti_nondon_form(): TForm() {};
  TEti_nondon_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TEti_nondon_form() {};
};

class TStampaNonDon : public TPrintapp
{
  static bool filter_func_nondon(const TRelation* rel);

  TRelation*        _rel;
  TMask*            _msk;
  TEti_nondon_form*	_form_eti;
  TAssoc_array			_categorie;
  int 							_cur;
  TParagraph_string _cognome_nome, _dencom;
  TDate 						_data_stampa;
  TDate							_data;
  ts								_tipostampa;
  TString16					_codsez, _codsot;
  int								_etlarghezza, _etcolonne;  
  bool							_stampa80;
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page (int file, int counter);

public:
  void crea_intestazione();
  void filtra_sezioni();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  TMask& app_mask() { return *_msk; }

  TStampaNonDon() : _data_stampa(TODAY), _cognome_nome("",25), _dencom("",50) {}
};
                                             
HIDDEN inline TStampaNonDon& app() { return (TStampaNonDon&) main_app(); }

TCursor* TEti_nondon_form::cursor() const { return app().current_cursor(); }

TRelation* TEti_nondon_form::relation() const { return cursor()->relation(); }

void TStampaNonDon::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  TRectype da(LF_SOGGETTI);
  TRectype a(LF_SOGGETTI);   
	if (sezini.not_empty())  
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

void TStampaNonDon::set_page(int file, int cnt)
{
	switch (_tipostampa)
	{
		case etichette:
		{ 
			TPrint_section& corpo = _form_eti->get_body();
			for (int r=1;r<=_etcolonne;r++)
			{                     
				if (current_cursor()->pos()<current_cursor()->items())
				{ 
					force_setpage(TRUE);
					corpo.update();
  				for (word i = 0; i < corpo.height(); i++)
  				{
  					TPrintrow& riga = corpo.row(i);
  					TString256 riga1 = riga.row();
						riga1.cut(_etlarghezza);
  					int colonna = ((r-1)*_etlarghezza);
  					TString16 formato;
  					formato << '@' << colonna << "g";
  					riga1.insert(formato,0);
  					set_row(i+1,riga1);
					}
					if (r < _etcolonne)  			
						++(*current_cursor());
				}					
			}				
		}
		break;
		case elenco:
		{
			if (_stampa80)
			{                     
				set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
				set_row(2,"@0g@8,rs", FLD(LF_SOGGETTI,SOG_TESSAVIS));
				set_row(3,"@6g@S", FLD(LF_SOGGETTI,SOG_CATDON));
				set_row(1,"@9g#a", &_cognome_nome);
				set_row(3,"@9g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
				set_row(1,"@44g@S", FLD(LF_SOGGETTI,SOG_TELABI));
				set_row(2,"@44g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
				set_row(3,"@44g@S", FLD(LF_SOGGETTI,SOG_TELALT));
				set_row(1,"@60g@ld", FLD(LF_SOGGETTI,SOG_DATAULTDON));
				set_row(2,"@64g@S", FLD(LF_SOGGETTI,SOG_TIPOULTDON));
				set_row(1,"@74g@n", FLD(LF_SOGGETTI,SOG_TOTDON));
				set_row(4,"");
			}			
			else
			{
				set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
				set_row(1,"@9g@S", FLD(LF_SOGGETTI,SOG_CATDON));
				set_row(1,"@12g#a", &_cognome_nome);
				set_row(1,"@38g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
				set_row(1,"@49g@S", FLD(LF_SOGGETTI,SOG_DOM_INDIR));
				set_row(1,"@100g@S", FLD(LF_SOGGETTI,SOG_TELABI));
				set_row(1,"@116g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
				set_row(1,"@132g@ld", FLD(LF_SOGGETTI,SOG_DATAULTDON));
				set_row(2,"@49g#a", &_dencom);
				set_row(2,"@116g@S", FLD(LF_SOGGETTI,SOG_TELALT));
				set_row(2,"@132g@S", FLD(LF_SOGGETTI,SOG_TIPOULTDON));
			}				
		}
		break;					
	}	
}

bool TStampaNonDon::filter_func_nondon(const TRelation* rel)
{   
	bool filtrato = TRUE;
	TLocalisamfile& sog = rel->lfile();
	// filtro per categorie
	TAssoc_array& categorie = app()._categorie;
	if (categorie.items() != 0)
	{
		const TString16 cat = sog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) cat);
	}	
	// filtro per data ultima donazione
	if (filtrato)		
	{
		const TDate dataultdon = sog.get(SOG_DATAULTDON);
		if (dataultdon.ok())
			filtrato = (dataultdon <= app()._data);
		else
			filtrato = FALSE;
  }  	
  return filtrato;
}

void TStampaNonDon::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(132);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if ((densot.ok())&& (densot.not_empty()))
	{
		intestazione << "/";
		intestazione << densot;
	}		
	if (_stampa80)
		intestazione.center_just(80);
	else		
		intestazione.center_just(132);		
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

bool TStampaNonDon::preprocess_page(int file, int counter)
{
	if (_tipostampa == elenco)
	{
		TRectype recsog = current_cursor()->curr();
		TString80 nome = recsog.get(SOG_COGNOME);
		nome << " ";
		nome << recsog.get(SOG_NOME);
		_cognome_nome = nome;
		TString256 localita = "";
		localita << current_cursor()->curr(-ALIAS_LCP).get("S6");
		if (localita.not_empty() && localita.ok())
		{
			localita << " ";
			localita << current_cursor()->curr(-ALIAS_LCP).get("S0");
			localita << " - ";		
		}
		else
		{
			localita = current_cursor()->curr(LF_COMUNI).get(COM_CAPCOM);
			localita << " ";
		}			
		localita << current_cursor()->curr(LF_COMUNI).get(COM_DENCOM);
		localita.trim();
		_dencom = localita;

		// salto pagina se cambio sezione
		const TString16 codsez = recsog.get(SOG_CODSEZ);		
		const TString16 codsot = recsog.get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if (_codsez != "**")
				printer().formfeed();
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
	}		
	if (_tipostampa==etichette)		
		if (printer().rows_left() < _form_eti->get_body().height())
			printer().formfeed();
  return TRUE;
}

bool TStampaNonDon::set_print(int m)
{
	_tipostampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  	_tipostampa = elenco;
  	_codsez = "**";
  	_codsot = "**";
  	break;
  case F_ETICHETTE:
  	_tipostampa = etichette;	
  	break;
  }	
	if (_tipostampa != undefined)
  {
    reset_files(); 
    add_file(LF_SOGGETTI);
		filtra_sezioni();
    _data = _msk->get(F_DATA);
	  const TString16 catpri = _msk->get(F_CAT1);
	  const TString16 catsec = _msk->get(F_CAT2);
	  const TString16 catter = _msk->get(F_CAT3);
	  const TString16 catqua = _msk->get(F_CAT4);
	  const TString16 catqui = _msk->get(F_CAT5);
	  const TString16 catses = _msk->get(F_CAT6);
	  if (catpri.not_empty() && catpri.ok())
	  	_categorie.add((const char*) catpri);
	  if (catsec.not_empty() && catsec.ok())
	  	_categorie.add((const char*) catsec);
	  if (catter.not_empty() && catter.ok())
	  	_categorie.add((const char*) catter);
	  if (catqua.not_empty() && catqua.ok())
	  	_categorie.add((const char*) catqua);
	  if (catqui.not_empty() && catqui.ok())
	  	_categorie.add((const char*) catqui);
	  if (catses.not_empty() && catses.ok())
	  	_categorie.add((const char*) catses);
		current_cursor()->set_filterfunction (filter_func_nondon);
		reset_print();
    crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TStampaNonDon::crea_intestazione()
{
  reset_header();
  if (_tipostampa == elenco)
  {
  	TString sep(132);
  	sep = "ELENCO SOGGETTI CHE NON DONANO DAL ";
  	if (app()._data.ok())
  		sep << _data.string();
		if (_stampa80)
			sep.center_just(80);  		
		else			
			sep.center_just(132);
  	set_header(2, "@0g%s", (const char*) sep);
  	TString16 data_stampa = _data_stampa.string();
  	set_header(2,"@0g%10s", (const char*) data_stampa);
  	sep = "";
  	sep << "Pag. @#";
  	if (_stampa80)
  		set_header(2, "@73g%s", (const char*) sep);  
		else  		
			set_header(2, "@110g%s", (const char*) sep);  		
  	set_header(3, "");
		if (_stampa80)
		{  	
  		set_header(4,"@0gCodice@9gCognome e nome@44gTelefono abit.@60gUltima don.@71gTot.don.");
  		set_header(5,"@0gTessera@44gTelefono lavoro");
  		set_header(6,"@0gCateg.@9gNato il@44gTelefono altro");
  		set_header(7,"@0g--------@9g----------------------------------@44g--------------@60g----------@71g--------");
			_cognome_nome.set_width(35);
  		
  	}
  	else
  	{
  		set_header(4,"@0gCodice@9gC.@12gCognome e nome@38gNato il@49gIndirizzo@100gTelefono abit.@116gTelefono lavoro@132gData u.d.");
  		set_header(5,"@49gCAP/Localit/Comune/Prov.@116gTelefono altro@132gTipo u.d.");
  		set_header(6,"@0g--------@9g--@12g-------------------------@38g----------@49g--------------------------------------------------@100g---------------@116g---------------@132g----------");
			_cognome_nome.set_width(25);
		}  		
	}  	
}

bool TStampaNonDon::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCP);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM");
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");
  _cur = add_cursor(new TCursor(_rel, "", 3));
  _msk = new TMask("at2800a");

  TConfig config(CONFIG_STUDIO);
  TString16 etformato = config.get("EtFormato");
  _etlarghezza = config.get_int("EtLarghezza");
  _etcolonne = config.get_int("EtColonne");
	_form_eti = new TEti_nondon_form(etformato);
  _stampa80 = config.get_bool("Stampa80");
  return TRUE;
}

bool TStampaNonDon::user_destroy()
{
  delete _msk;
  delete _rel;
  delete _form_eti;
  return TRUE;
}

int at2800(int argc, char* argv[])
{
  TStampaNonDon a;
  a.run(argc, argv, "Stampa soggetti che non donano dal ...");
  return 0;
}
