#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <tabutil.h>
#include <utility.h>

#include "soggetti.h"
#include "sezioni.h"
#include "convoc.h"
#include "rconvoc.h"

#include "at4.h"
#include "at4400a.h"
#include "at4100b.h"
#include "at4100c.h"

#define ALIAS_LCP	100
#define ALIAS_LDN	400

enum ts { undefined = 0, elenco = 1, etichette = 2, cartoline = 3 };

// definizione form per etichette
class TStConvoc_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TStConvoc_form(): TForm() {};
  TStConvoc_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TStConvoc_form() {};
};

class TConvocazioniPerSezione : public TPrintapp
{
  TRelation*				_rel;
  TMask*      			_msk;
  TStConvoc_form*		_form_eti;
  TStConvoc_form*   _form_car;
  int 							_cur;
  TParagraph_string _cognome_nome;
  TDate 						_data_stampa;
  TDate							_dataini, _datafin;
  bool 							_aggiorna;
	TString16 				_punto,_tipo, _sezione, _sottog, _ordina;
	TString16					_codsez, _codsot;
	TString16					_orario, _datacart;
	TString80					_invitoper, _data, _presso1, _presso2, _presso3, _presso4;
	TString80					_note, _intest1, _intest2, _intest3, _intest4;
	bool 							_usasez;
	ts								_tipostampa;
	int 							_contxsez;
	int								_etlarghezza, _etcolonne;

  //static bool filter_func_stconvoc(const TRelation* rel);
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page(int file, int counter);
  virtual print_action postprocess_print(int file, int counter);
  ts dati_cartoline();

public:
  void crea_intestazione();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void footer_sezione();
  void dati_sezione(const TString16 codsez, const TString16 codsot);
  TMask& app_mask() { return *_msk; }
  TConvocazioniPerSezione() : _data_stampa(TODAY), _cognome_nome("",35) {}
};

HIDDEN inline TConvocazioniPerSezione& app() { return (TConvocazioniPerSezione&) main_app(); }

TCursor* TStConvoc_form::cursor() const { return app().current_cursor(); }

TRelation* TStConvoc_form::relation() const { return cursor()->relation(); }

ts TConvocazioniPerSezione::dati_cartoline()
{
	TMask msk("at4100b");
 	if (msk.run() == K_ENTER) 
 	{ 
 		_invitoper = msk.get(F_INVITOPER);
 		_data			 = msk.get(F_DATA);
 		_presso1   = msk.get(F_PRESSO1);
 		_presso2   = msk.get(F_PRESSO2);
 		_presso3   = msk.get(F_PRESSO3);
 		_presso4   = msk.get(F_PRESSO4);
 		_datacart  = msk.get(F_DATACART);
 		_orario  	 = msk.get(F_ORARIO);
 		_note   	 = msk.get(F_NOTE);
 		_usasez		 = msk.get_bool(F_USASEZ);
 		return cartoline;
 	}
 	else
 		return undefined;
}

void TConvocazioniPerSezione::set_page(int file, int cnt)
{
	// costruzione etichette
	switch (_tipostampa)
	{
		case etichette:
		{
			TPrint_section& corpo = _form_eti->get_body();
			for (int r=1;r<=_etcolonne;r++)
			{                     
				if (current_cursor()->pos()<current_cursor()->items())
				{ 
					force_setpage(TRUE);
					corpo.update();
  				for (word i = 0; i < corpo.height(); i++)
  				{
  					TPrintrow& riga = corpo.row(i);
  					TString256 riga1 = riga.row();
						riga1.cut(_etlarghezza);
  					int colonna = ((r-1)*_etlarghezza);
  					TString16 formato;
  					formato << '@' << colonna << "g";
  					riga1.insert(formato,0);
  					set_row(i+1,riga1);
					}
					if (r < _etcolonne)  			
						++(*current_cursor());
				}					
			}				
		}
		break;
		case cartoline:
		{
			TPrint_section& corpo = _form_car->get_body();
			const TString16 codsez = current_cursor()->curr().get(SOG_CODSEZ);		
			const TString16 codsot = current_cursor()->curr().get(SOG_CODSOT);		
			if ((_codsez!=codsez)||(_codsot!=codsot))
			{
				_codsez = codsez;
				_codsot = codsot;
				dati_sezione(codsez, codsot);
				corpo.reset();
			  TForm_item& intest1 = corpo.find_field(CAR_INTEST1);
			  intest1.set(_intest1);
			  TForm_item& intest2 = corpo.find_field(CAR_INTEST2);
			  intest2.set(_intest2);
			  TForm_item& intest3 = corpo.find_field(CAR_INTEST3);
			  intest3.set(_intest3);
			  TForm_item& intest4 = corpo.find_field(CAR_INTEST4);
			  intest4.set(_intest4);
			  TForm_item& invito = corpo.find_field(CAR_INVITOPER);
	  		invito.set(_invitoper);
			  TForm_item& data = corpo.find_field(CAR_DATA);
	  		data.set(_data);
			  TForm_item& presso1 = corpo.find_field(CAR_PRESSO1);
	  		presso1.set(_presso1);
			  TForm_item& presso2 = corpo.find_field(CAR_PRESSO2);
	  		presso2.set(_presso2);
			  TForm_item& presso3 = corpo.find_field(CAR_PRESSO3);
	  		presso3.set(_presso3);
			  TForm_item& presso4 = corpo.find_field(CAR_PRESSO4);
	  		presso4.set(_presso4);
			  TForm_item& orario = corpo.find_field(CAR_ORARIO);
	  		orario.set(_orario);
			  TForm_item& datacart = corpo.find_field(CAR_DATACART);
	  		datacart.set(_datacart);
			  TForm_item& note = corpo.find_field(CAR_NOTE);
	  		note.set(_note);
			}
			corpo.update();
  		for (word i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case elenco:
		{
			set_row(1,"@0g#a", &_cognome_nome);
			set_row(1,"@36g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			set_row(1,"@40g@S", FLD(LF_SOGGETTI,SOG_TESSAVIS));
			set_row(1,"@48g@ld", FLD(LF_SOGGETTI,SOG_DATAULTDON));
			set_row(1,"@59g@S", FLD(LF_SOGGETTI,SOG_TIPOULTDON));
			set_row(1,"@62g@ld", FLD(LF_RCONVOC,RCV_DATACONV));
			set_row(1,"@73g@S", FLD(LF_RCONVOC,RCV_CHIAMATA));
		}
		break;
	}					
}

print_action TConvocazioniPerSezione::postprocess_print(int file, int counter)
{ 
	if (_contxsez > 0)
		footer_sezione();
	return NEXT_PAGE;
}

bool TConvocazioniPerSezione::preprocess_page(int file, int counter)
{
	if (_tipostampa==elenco)
	{
		TString80 nome = current_cursor()->curr(LF_SOGGETTI).get(SOG_COGNOME);
		nome << " ";
		nome << current_cursor()->curr(LF_SOGGETTI).get(SOG_NOME);
		_cognome_nome = nome;

		// salto pagina se cambio sezione
		const TString16 codsez = current_cursor()->curr(LF_SOGGETTI).get(SOG_CODSEZ);		
		const TString16 codsot = current_cursor()->curr(LF_SOGGETTI).get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if (_codsez != "**")
				footer_sezione();
			_contxsez = 0;				
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
		_contxsez++;
	}
	if (_tipostampa==etichette)		
		if (printer().rows_left() < _form_eti->get_body().height())
			printer().formfeed();
  return TRUE;
}

void TConvocazioniPerSezione::footer_sezione()
{
	// stampa totale sezione
	reset_footer();
	TString sep(80);
	sep.fill('-');
	set_footer(2, (const char *) sep);
	set_footer(3,"TOTALE CONVOCATI SEZIONE    %d", _contxsez);
	printer().formfeed(); 
	reset_footer();
}

void TConvocazioniPerSezione::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(80);
	intestazione = "Sezione: ";
	intestazione << codsez;
	if (codsot.not_empty())
	{
		intestazione << "/";
		intestazione << codsot;
	}		
	intestazione << " ";
	intestazione << densez;
	if (densot.not_empty())
	{
		intestazione << "/";
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
 	TString16 data_stampa = _data_stampa.string();
 	set_header(1,"@0g%10s", (const char*) data_stampa);
 	TString16 sep = "";
 	sep << "Pag. @#";
 	set_header(1, "@70g%s", (const char*) sep);
	return;
}

void TConvocazioniPerSezione::dati_sezione(const TString16 codsez, const TString16 codsot)
{
	_intest1 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ1);
	_intest2 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ2);
	_intest3 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ3);
	_intest4 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ4);
	if (_usasez)
	{
		if (_presso1.blank())
		{
			_presso1 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE1);
			_presso2 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE2);
			_presso3 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE3);
			_presso4 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE4);
		}			
		if (_orario.blank())
			_orario = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVORA);
		if (_note.blank())
			_note = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVNOTE);
	}
	return;
}

bool TConvocazioniPerSezione::set_print(int m)
{ 
	_tipostampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  	_tipostampa = elenco;
  break;
  case F_ETICHETTE:
  	_tipostampa = etichette;	
	break;  	
  case F_CARTOLINE:
  	_tipostampa = dati_cartoline();	
  break;
  }	
	if (_tipostampa != undefined)
  {
    _dataini 		= _msk->get(F_DATAINI);
    _datafin 		= _msk->get(F_DATAFIN);
    _punto      = _msk->get(F_PUNTO);
    _tipo				= _msk->get(F_TIPO);
    _codsez     = _msk->get(F_CODSEZ);
    _codsot			= _msk->get(F_CODSOT);
    _aggiorna   = _msk->get_bool(F_AGGIORNA);
    _ordina     = _msk->get(F_ORDINA);
		TString256 chiave = "";
		switch (_ordina[0])
		{
			case 'A':
				chiave = "90->CODSEZ|90->CODSOT|UPPER(90->COGNOME)|UPPER(90->NOME)|116->DATACONV";
			break;
			case 'C':
				chiave = "90->CODSEZ|90->CODSOT|90->DOM_CAP|UPPER(90->COGNOME)|UPPER(90->NOME)|116->DATACONV";
			break;
		}
		TString80 filtro = "";			
  	// filtro per punto di raccolta
  	if (_punto.not_empty())
  	{
  		if (filtro.not_empty())
				filtro << " && ";
			filtro << format("(115->PUNTO == \"%s\")",(const char*)_punto);
		}	
		// filtro per tipo
		if (_tipo.not_empty())
		{
  		if (filtro.not_empty())
				filtro << " && ";
			filtro << format("(115->TIPO == \"%s\")",(const char*)_tipo);
		}			
		// filtro per sezione/sottogruppo
		if (_sezione.not_empty())
		{
  		if (filtro.not_empty())
				filtro << " && ";
 			filtro << format("(90->CODSEZ == \"%s\")",(const char*)_sezione);
			if (_sottog.not_empty())
			{
	  		if (filtro.not_empty())
					filtro  << " && ";
				filtro << format("(90->CODSOT == \"%s\")",(const char*)_sottog);
			}
		}
		// filtro per data
	  TRectype da(LF_RCONVOC);
  	TRectype a (LF_RCONVOC);
  	if (_dataini.ok())
			da.put(RCV_DATACONV, _dataini);
  	if (_datafin.ok())
			a.put(RCV_DATACONV, _datafin);
	  _cur = add_cursor(new TSorted_cursor(_rel, (const char*) chiave,"", 2, &da, &a));
		current_cursor()->setfilter((const char*) filtro, TRUE);
		//current_cursor()->set_filterfunction(filter_func_stconvoc);
  	_codsez = "**";
  	_codsot = "**";
    reset_files(); 
    add_file(LF_RCONVOC);
		reset_print();
		crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TConvocazioniPerSezione::crea_intestazione()
{
  reset_header();
  if (_tipostampa == elenco)
  {
  	TString sep(80);
  	sep = "ELENCO CONVOCATI";
  	if (_tipo.not_empty())
  	{
  		sep << " per ";
  		sep << _tipo;
		}  		
		if (_dataini.ok())
		{
  		sep << " dal ";
 			sep << _dataini.string();
		}
		if (_datafin.ok())
		{ 			
 			sep << " al ";
 			sep << _datafin.string();
		}
		sep.center_just();
  	set_header(2, "@0g%s", (const char*) sep);
	 	char ora[16];
 		_strtime(ora);
 		set_header(2,"@0g%8s", ora);
		sep = "";
		if (_punto.not_empty())
		{         
 			sep << "presso ";
 			sep << _punto;
 			sep << " ";
 			TTable ldn("LDN"); 
			ldn.put("CODTAB",_punto);
			if (ldn.read() == NOERR)
				sep << ldn.get("S0");
		} 			
		sep.center_just();
  	set_header(3, "@0g%s", (const char*) sep);
	  set_header(5,"@0gCognome e nome@36gCat.@40gTessera@48gUltima donaz.@62gData conv.@73gTipo");
	  set_header(6,"@0g-----------------------------------@36g---@40g-------@48g---------- --@62g----------@73g----");
  	printer().footerlen(3);
	}  	
}

bool TConvocazioniPerSezione::user_create()
{
  _rel = new TRelation(LF_RCONVOC);
  _rel->add(LF_SOGGETTI, "CODICE==CODICE");
  _rel->add(LF_CONVOC, "NUMERO==NUMERO");
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,LF_SOGGETTI,ALIAS_LCP);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM",1,LF_SOGGETTI);
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT",1,LF_SOGGETTI);
  _msk = new TMask("at4400a");
  TConfig config(CONFIG_STUDIO);
  TString16 etformato = config.get("EtFormato");
  _etlarghezza = config.get_int("EtLarghezza");
  _etcolonne = config.get_int("EtColonne");
  _form_eti = new TStConvoc_form(etformato);
  _form_car = new TStConvoc_form("ATCARTO1");
  return TRUE;
}

bool TConvocazioniPerSezione::user_destroy()
{
  delete _msk;
  delete _rel;
  delete _form_eti;
  delete _form_car;
  return TRUE;
}

int at4400(int argc, char* argv[])
{
  TConvocazioniPerSezione a;
  a.run(argc, argv, "Elenco convocati");
  return 0;
}
