#include <form.h>
#include <mask.h>
#include <msksheet.h>
#include <printapp.h>
#include <utility.h>

#include "soggetti.h"
#include "sezioni.h"
#include <comuni.h>

#include "at4.h"
#include "at4600a.h"

#include "at4500.h"
#define ALIAS_COMDOM 501
#define ALIAS_COMNAS 500 
#define ALIAS_LCP    100
 
// definizione form per tessere associative
class TTessereS_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TTessereS_form(): TForm() {};
  TTessereS_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TTessereS_form() {};
};

class TStampaTessereS : public TPrintapp
{
  TRelation*			_rel;
  TIsamtempfile*	_sogtmp;
  TMask*      		_msk;
  TTessereS_form*	_form_pag;
  TAssoc_array		_asoggetti;
  int							_numdon;
  bool						_aggiorna;
  int 						_cur;
  TDate 					_data_stampa;
	TString					_riepilogodon;

protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page(int file, int counter);
	
	static void add_rows_soggetti(TSheet_field& s, int count = 10, int start = 1);
	static bool soggetti_notify(TSheet_field& s, int r, KEY k);
	static bool nome_handler(TMask_field& f, KEY k);
	static bool codice_handler(TMask_field& f, KEY k);
	
public:
	void filtra_codici();	
  TMask& app_mask() { return *_msk; }
  TStampaTessereS() : _data_stampa(TODAY), _riepilogodon(35) {}
};

HIDDEN inline TStampaTessereS& app() { return (TStampaTessereS&) main_app(); }

TCursor* TTessereS_form::cursor() const { return app().current_cursor(); }

TRelation* TTessereS_form::relation() const { return cursor()->relation(); }

HIDDEN bool printer_handler(TMask_field& f, KEY k)
{        
  if (k == K_SPACE)
  {                  
    TMask& m = f.mask();
    if (!m.query_mode())
    {
      TString16 config;
      config << "TESSERE";
  
      TPrinter& p = printer();
      p.set_printtype(normprinter);           // Force configuration update
      p.read_configuration(config);
      if (p.set())
      	f.message_box("Stampante configurata per stampa tessere");
    }    
  }
  return TRUE;  
}

void TStampaTessereS::add_rows_soggetti(TSheet_field& s, int count, int start)
{
	if (start == 1)
		s.destroy();
	for (int r=start; r<start+count; r++)
		TToken_string& row = s.row(r-1);
}

void TStampaTessereS::filtra_codici()
{
	TSheet_field& s = (TSheet_field&)_msk->field(F_SOGGETTI);
	for (int r=0; r < s.items(); r++)
 	{
 		TToken_string& row = s.row(r);
 		const long codice = row.get_long(0);
 		if (codice != 0)
 		{
 			TLocalisamfile soggetti(LF_SOGGETTI);
 			soggetti.setkey(1);
 			soggetti.zero();
 			soggetti.put(SOG_CODICE,codice);
 			if (soggetti.read() == NOERR)
 				_sogtmp->write(soggetti.curr());
 		}
	} 		
}	

void TStampaTessereS::set_page(int file, int cnt)
{
	TPrint_section& corpo = _form_pag->get_body();
	corpo.update();
  //TForm_item& rigadon = corpo.find_field(TES_RIGADON1);
	//rigadon.set(_riepilogodon);
	for (word i = 0; i < corpo.height(); i++)
	{
		TPrintrow& riga = corpo.row(i);
		set_row(i+1,riga);
	}  			
	force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
												// altrimenti stampa sempre la stessa etichetta
}

bool TStampaTessereS::preprocess_page(int file, int counter)
{                              
	TLocalisamfile& sogg = current_cursor()->file(LF_SOGGETTI);
	
	const int totdon = sogg.get_int(SOG_TOTDON);
	const TDate dataultima = sogg.get(SOG_DATAULTDON);
	_riepilogodon = "";
	if (totdon != 0)	
	{	
		_riepilogodon = "Donazioni fino al ";
		_riepilogodon << dataultima.string();
		_riepilogodon << " n.";
		_riepilogodon << totdon;
	}	
	if (_aggiorna)
	{
		sogg.put(SOG_T_STAMPATA,TRUE);
		sogg.rewrite();
	}	
	current_cursor()->file().remove();	
  return TRUE;
}

bool TStampaTessereS::set_print(int m)
{ 
	TSheet_field& s = (TSheet_field&)_msk->field(F_SOGGETTI);
	s.destroy();
	s.force_update();
	if (s.items()==0)
		add_rows_soggetti(s,50);
  KEY tasto;
  tasto = _msk->run();
  if (tasto == K_ENTER)
  { 
    reset_files(); 
    add_file(LF_SOGGETTI);
	  _asoggetti.destroy();
	 	for (int r=0; r < s.items(); r++)
	 	{
	 		TToken_string& row = s.row(r);
	 		const long codice = row.get_long(0);
	 		const char* scodice = row.get(0);
	 		if (codice != 0)
	 			_asoggetti.add(scodice);
	 	}
	 	if (_asoggetti.items() != 0)
	 	{
			_aggiorna = _msk->get_bool(F_AGGIORNA);
			filtra_codici();
		}				
		reset_print();
    return TRUE;
  }
  else
    return FALSE;
}

bool TStampaTessereS::user_create()
{
  _sogtmp = new TIsamtempfile(LF_SOGGETTI,NULL,TRUE,TRUE);
  _rel = new TRelation(_sogtmp);
	_rel->add(LF_SOGGETTI, "CODICE==CODICE");
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM",1,0,ALIAS_COMDOM);
  _rel->add(LF_COMUNI, "COM==COMNASC",1,0,ALIAS_COMNAS);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCP);
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");
	//cursore ordinamento per sezione+sottogruppo+cognome+nome  
  _cur = add_cursor(new TCursor(_rel, "", 1));
  _msk = new TMask("at4600a");
  _msk->set_handler(F_PRINTER, printer_handler);
  _form_pag = new TTessereS_form("ATTESSER");
	TSheet_field& ss = (TSheet_field&)_msk->field(F_SOGGETTI);
	ss.set_notify(soggetti_notify);
	ss.sheet_mask().set_handler(F_S_NOME,nome_handler);
	ss.sheet_mask().set_handler(F_S_CODICE,codice_handler);
	TString16   config;
	config << "TESSERE";
  printer().read_configuration(config);
  return TRUE;
}

bool TStampaTessereS::soggetti_notify(TSheet_field& s, int r, KEY k)
{              
	bool ok = TRUE;
	switch (k)
	{ 
		case K_INS:
			// richiesta inserimento riga
		break;	
		case K_DEL:
		case K_CTRL+K_DEL:
		// avvenuta cancellazione riga
		break;
		case K_SPACE:
		// inizio modifica riga
		break;
		case K_TAB:
		// entrata riga			
		{         
		  static bool entering = TRUE;
		  if (entering)
		  {
		    entering = FALSE;
			  if ((r == s.items()-1) && (r == s.first_empty()))
				  app().add_rows_soggetti(s,10,r+1);
			  TToken_string& row = s.row(r);
			  if (row.empty_items())
			  	s.select(r);			
			  entering = TRUE;	
			}  	
		}
	  break;
		case K_ENTER:
		// uscita da riga modificata
		case K_CTRL+K_TAB:
		// uscita riga
		{ 
			TToken_string& row = s.row(r);
			const long codsog = row.get_long(0);
			if (codsog != 0)
				for (int i=s.items()-1; i>=0; i--)
				{
					if (i != r)
					{
						TToken_string& row = s.row(i); 
						if (codsog == row.get_long(0))
							return s.sheet_mask().field(F_S_CODICE).error_box("Soggetto gia' inserito");
					}					
				}				
			else
			{
				const char* cognome = row.get(1);
				if ((cognome != NULL) && (cognome != "\0"))
					s.sheet_mask().field(F_S_NOME).set_focusdirty();
			}
		}
	  break;  	
	}  
  return ok;  
}

bool TStampaTessereS::nome_handler(TMask_field& f, KEY k)
{
	bool ok = TRUE;
  if (f.to_check(k))
  {
  	TMask& m = f.mask();
  	long codsog = m.get_long(F_S_CODICE);
  	if (codsog == 0)
      f.on_key(K_F9);
  }	
  return ok;
}

bool TStampaTessereS::codice_handler(TMask_field& f, KEY k)
{
	bool ok = TRUE;
  if (f.to_check(k))
  {
  	TMask& m = f.mask();
  	long codsog = m.get_long(F_S_CODICE);
  	if (codsog != 0)
  	{
			TLocalisamfile sog(LF_SOGGETTI);
  		sog.setkey(1);	
			sog.zero();
			sog.put(SOG_CODICE, codsog);
			int err = sog.read();
			if (err == NOERR)
			{
  			m.set(F_S_COGNOME, sog.get(SOG_COGNOME));
  			m.set(F_S_NOME, sog.get(SOG_NOME));
  			m.set(F_S_DATANASC, sog.get(SOG_DATANASC));
  			m.set(F_S_TESSAVIS, sog.get(SOG_TESSAVIS));
  			m.set(F_S_CODSEZ, sog.get(SOG_CODSEZ));
  			m.set(F_S_CODSOT, sog.get(SOG_CODSOT));
  			m.set(F_S_CATDON, sog.get(SOG_CATDON));
			}  			
			else
				ok = FALSE; // codice non esistente  			
  	}
  }	
  return ok;
}

bool TStampaTessereS::user_destroy()
{
  delete _sogtmp;
  delete _msk;
  delete _rel;
  delete _form_pag;
  printer().read_configuration();
  return TRUE;
}

int at4600(int argc, char* argv[])
{
  TStampaTessereS a;
  a.run(argc, argv, "Stampa tessere singole");
  return 0;
}
